// Copyright (C) 2015 Tuma Solutions, LLC
// Process Dashboard - Data Automation Tool for high-maturity processes
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, see <http://www.gnu.org/licenses/>.
//
// The author(s) may be contacted at:
//     processdash@tuma-solutions.com
//     processdash-devel@lists.sourceforge.net

package net.sourceforge.processdash.ui.help;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.net.URL;
import java.security.AccessController;
import java.security.PrivilegedAction;

import javax.help.JHelpContentViewer;
import javax.help.plaf.basic.BasicContentViewerUI;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.Timer;
import javax.swing.event.HyperlinkEvent;
import javax.swing.plaf.ComponentUI;

import net.sourceforge.processdash.ui.Browser;

public class DashHelpContentViewerUI extends BasicContentViewerUI {

    public static ComponentUI createUI(JComponent x) {
        return new DashHelpContentViewerUI((JHelpContentViewer) x);
    }



    private Timer anchorTimer;

    private URL activatedUrl;

    private JEditorPane html;

    private int scrollCount;

    public DashHelpContentViewerUI(JHelpContentViewer viewer) {
        super(viewer);
        anchorTimer = new Timer(ANCHOR_RESET_DELAY, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                scrollToAnchor();
            }
        });
    }


    /**
     * When the user clicks on a hyperlink to an external website, open the URL
     * in a real web browser instead of the JavaHelp window.
     */
    @Override
    public void hyperlinkUpdate(HyperlinkEvent he) {
        if (he.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
            activatedUrl = he.getURL();
            String p = activatedUrl.getProtocol();
            if (p.equalsIgnoreCase("http") || p.equalsIgnoreCase("https")) {
                AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        Browser.launch(activatedUrl.toString());
                        return null;
                    }
                });
                return;
            }
        }
        super.hyperlinkUpdate(he);
    }


    /**
     * Asynchronous loading of images in the JEditorPane causes the page to
     * scroll. That foils the ability to jump to a named anchor in an HTML
     * document. This method attempts to fix that problem by:
     * <ol>
     * <li>Watching for the page load event</li>
     * <li>If the new page is for a URL with an anchor, starts a timer</li>
     * <li>The timer fires repeatedly and resets the scroll location to the
     * named anchor</li>
     * </ol>
     */
    @Override
    public void propertyChange(PropertyChangeEvent evt) {
        super.propertyChange(evt);

        if (evt.getSource() instanceof JEditorPane
                && "page".equals(evt.getPropertyName())) {
            html = (JEditorPane) evt.getSource();
            URL url = html.getPage();
            if (url != null && url.equals(activatedUrl) && url.getRef() != null) {
                scrollCount = ANCHOR_RESET_TIME / ANCHOR_RESET_DELAY;
                anchorTimer.start();
            }
        }
    }

    private void scrollToAnchor() {
        URL url = html.getPage();
        if (url != null && url.equals(activatedUrl) && url.getRef() != null
                && scrollCount-- > 0) {
            html.scrollToReference(url.getRef());
        } else {
            activatedUrl = null;
            scrollCount = 0;
            anchorTimer.stop();
        }
    }

    /** Milliseconds to wait between resets of the anchor scroll location */
    private static final int ANCHOR_RESET_DELAY = 100;

    /** Length of time to spend repeatedly resetting the scroll location */
    private static final int ANCHOR_RESET_TIME = 1200;

}
